// iPaq 3D engine  -  Janurary 2001  -  The Phantom
// Minimal 3D engine for Handhelds
// ------------------------------------------------

#include "rasterize.h"
#include "engine.h"
#include "string.h"

// ------------------------------------------------
// ARasterizer::ARasterizer
// Constructor
// ------------------------------------------------
ARasterizer::ARasterizer()
{
	m_Dest = 0;
	m_ZBuffer = new PIXEL[240 * 320];
	for ( int i = 0; i < 240; i++ )
	{
		m_Left[i] = IFP(319);
		m_Right[i] = 0;
	}
	m_Lines = 0;
}

// ------------------------------------------------
// ARasterizer::~ARasterizer
// Destructor
// ------------------------------------------------
ARasterizer::~ARasterizer()
{
}

// ------------------------------------------------
// ARasterizer::nextframe
// Update stuff for a new frame
// ------------------------------------------------
void ARasterizer::nextframe()
{
#ifdef INTERLACE
	if (m_Lines) m_Lines = 0; else m_Lines = 1;
	unsigned long addr = m_Lines * 320;
	for ( int i = m_Lines; i < 240; i+= 2 ) 
	{
		memset( m_Dest + addr, 0, 640 );
		memset( m_ZBuffer + addr, 65535, 640 );
		addr += 640;
	}
#else
	memset( m_Dest, 0, 320 * 240 * 2 );
	memset( m_ZBuffer, 0, 320 * 240 * 2 );	
#endif
}

// ------------------------------------------------
// ARasterizer::update
// Copy the frame buffer to the display
// ------------------------------------------------
void ARasterizer::update( PIXEL* a_Dest )
{
#ifdef INTERLACE
	unsigned long addr = m_Lines * 320;
	for ( int i = m_Lines; i < 240; i+= 2 )
	{
		memcpy( a_Dest + addr, m_Dest + addr, 640 );
		addr += 640;
	}
#else
	memcpy( a_Dest, m_Dest, 640 * 240 );
#endif
}

// ------------------------------------------------
// ARasterizer::draw
// Draw a polygon using outline tables
// ------------------------------------------------
void ARasterizer::draw( APolygon* a_Poly )
{
	int miny = 239;
	int maxy = 0;
	int sx1 = a_Poly->getvertex( 0 )->getcoord()->getscreenx();
	int sy1 = a_Poly->getvertex( 0 )->getcoord()->getscreeny();
	int u1 = a_Poly->getvertex( 0 )->getu1();
	int v1 = a_Poly->getvertex( 0 )->getv1();
	int z1 = (0x7FFFFFF0 / a_Poly->getvertex( 0 )->getcoord()->getrot().getz()) << FPP;
	for ( int i = 0; i < 3; i++ )
	{
		int sx2 = a_Poly->getvertex( (i + 1) % 3 )->getcoord()->getscreenx();
		int sy2 = a_Poly->getvertex( (i + 1) % 3 )->getcoord()->getscreeny();
		int u2 = a_Poly->getvertex( (i + 1) % 3 )->getu1();
		int v2 = a_Poly->getvertex( (i + 1) % 3 )->getv1();
		int z2 = (0x7FFFFFF0 / a_Poly->getvertex( (i + 1) % 3 )->getcoord()->getrot().getz()) << FPP;
		int du, dv, dz, dx, cu, cv, cz, cx, cy, y1, y2;
		if (sy2 > sy1)
		{
			cx = sx1;
			cy = sy1;
			cu = u1;
			cv = v1;
			cz = z1;
			y1 = (sy1 >> FPP) + 1;
			y2 = (sy2 >> FPP) + 1;
			if (y1 < miny) miny = y1;
			if (y2 > maxy) maxy = y2;
			int d = sy2 - sy1;
			if ((d > 4) || (d < -4)) 
			{
				dx = FPDIV( sx2 - sx1, d ); 
				du = FPDIV( u2 - u1, d ); 
				dv = FPDIV( v2 - v1, d ); 
				dz = FPDIV( z2 - z1, d ); 
			} else du = dv = dx = 0;
			int fix = FP_ONE - (sy1 & (FP_ONE - 1));
			cx += FPMUL( fix, dx );
			cu += FPMUL( fix, du );
			cv += FPMUL( fix, dv );
		}
		else
		{
			cx = sx2;
			cy = sy2;
			cu = u2;
			cv = v2;
			cz = z2;
			y1 = (sy2 >> FPP) + 1;
			y2 = (sy1 >> FPP) + 1;
			if (y2 < miny) miny = y2;
			if (y1 > maxy) maxy = y1;
			int d = sy1 - sy2;
			if ((d > 4) || (d < -4)) 
			{
				dx = FPDIV( sx1 - sx2, d ); 
				du = FPDIV( u1 - u2, d ); 
				dv = FPDIV( v1 - v2, d ); 
				dz = FPDIV( z1 - z2, d ); 
			} else du = dv = dx = 0;
			int fix = FP_ONE - (sy2 & (FP_ONE - 1));
			cx += FPMUL( fix, dx );
			cu += FPMUL( fix, du );
			cv += FPMUL( fix, dv );
		}
		for ( int i = y1; i < y2; i++ )
		{
			if (cx < m_Left[i]) 
			{
				m_Left[i] = cx;
				m_ULeft[i] = cu;
				m_VLeft[i] = cv;
				m_ZLeft[i] = cz;
			}
			if (cx > m_Right[i]) 
			{
				m_Right[i] = cx;
				m_URight[i] = cu;
				m_VRight[i] = cv;
				m_ZRight[i] = cz;
			}
			cx += dx;
			cu += du;
			cv += dv;
			cz += dz;
		}
		sx1 = sx2;
		sy1 = sy2;
		u1 = u2;
		v1 = v2;
		z1 = z2;
	}
	PIXEL* dst = m_Dest + miny * 320;
	PIXEL* src = a_Poly->gettexture()->getbitmap();
	PIXEL* zbf = m_ZBuffer + miny * 320;
	for ( i = miny; i < maxy; i++ )
	{
		int x1 = m_Left[i] >> FPP, x2 = m_Right[i] >> FPP;
		if (x2 > x1)
		{
			int du = FPDIV( m_URight[i] - m_ULeft[i], m_Right[i] - m_Left[i] );
			int dv = FPDIV( m_VRight[i] - m_VLeft[i], m_Right[i] - m_Left[i] );
			int dz = ((m_ZRight[i] - m_ZLeft[i]) / (m_Right[i] - m_Left[i])) >> (13 - FPP);
			unsigned long cuv = ((m_ULeft[i] >> (FPP - 8)) << 16) + (m_VLeft[i] >> (FPP - 8));
			unsigned long duv = ((du >> (FPP - 8)) << 16) + (dv >> (FPP - 8));
			unsigned short cz = m_ZLeft[i] >> 13;
			for ( int j = x1; j < x2; j++ ) 
			{
				if (*(zbf + j) < cz)
				{
					*(dst + j) = *(src + (cuv >> 24) + (cuv & 0xFF00));
					*(zbf + j) = cz;
				}
				cuv += duv;
				cz += dz;
			}
		}
		dst += 320;
		zbf += 320;
		m_Left[i] = IFP(319);
		m_Right[i] = 0;
	}
}

// ------------------------------------------------
// ARasterizer::drawtri
// Faster triangle rasterizing
// ------------------------------------------------
void ARasterizer::drawtri( APolygon* a_Poly )
{
	AVertex* V1 = a_Poly->getvertex( 0 ); 
	AVertex* V2 = a_Poly->getvertex( 1 ); 
	AVertex* V3 = a_Poly->getvertex( 2 ); 
	if ( V1->getcoord()->getscreeny() > V2->getcoord()->getscreeny()) { AVertex* t = V1; V1 = V2; V2 = t; }
	if ( V1->getcoord()->getscreeny() > V3->getcoord()->getscreeny()) { AVertex* t = V1; V1 = V3; V3 = t; }
	if ( V2->getcoord()->getscreeny() > V3->getcoord()->getscreeny()) { AVertex* t = V2; V2 = V3; V3 = t; }
	ACoord* c1 = V1->getcoord();
	ACoord* c2 = V2->getcoord();
	ACoord* c3 = V3->getcoord();
	int x1 = c1->getscreenx(), y1 = c1->getscreeny(), z1 = (0x7FFFFFF0 / c1->getrot().getz()) << FPP, u1 = V1->getu1(), v1 = V1->getv1();
	int x2 = c2->getscreenx(), y2 = c2->getscreeny(), z2 = (0x7FFFFFF0 / c2->getrot().getz()) << FPP, u2 = V2->getu1(), v2 = V2->getv1();
	int x3 = c3->getscreenx(), y3 = c3->getscreeny(), z3 = (0x7FFFFFF0 / c3->getrot().getz()) << FPP, u3 = V3->getu1(), v3 = V3->getv1();
	int ldx, ldu, ldv, ldz, rdx, rdu, rdv, rdz, llen = y3 - y1, rlen = y2 - y1;
	if (llen == 0) return; else 
	{
		ldx = FPDIV( x3 - x1, llen );
		ldu = FPDIV( u3 - u1, llen );
		ldv = FPDIV( v3 - v1, llen );
		ldz = FPDIV( z3 - z1, llen );
	}
	if (rlen == 0) rdx = rdu = rdv = rdz = 0; else 
	{
		rdx = FPDIV( x2 - x1, rlen );
		rdu = FPDIV( u2 - u1, rlen );
		rdv = FPDIV( v2 - v1, rlen );
		rdz = FPDIV( z2 - z1, rlen );
	}
	int temp = FPDIV( rlen, llen );
	int d = FPMUL( temp, x3 - x1 ) + (x1 - x2 );
	bool lleft = true;
	if (d >= 0)
	{
		int temp = ldx; ldx = rdx; rdx = temp;
		temp = ldu; ldu = rdu; rdu = temp;
		temp = ldv; ldv = rdv; rdv = temp;
		temp = ldz; ldz = rdz; rdz = temp;
		temp = llen; llen = rlen; rlen = temp;
		lleft = false;
	}
	int diff = (FP_ONE - 1) - (y1 & (FP_ONE - 1));
	int lx = x1 + FPMUL( diff, ldx ), rx = x1 + FPMUL( diff, rdx );
	int lu = u1 + FPMUL( diff, ldu ), ru = u1 + FPMUL( diff, rdu );
	int lv = v1 + FPMUL( diff, ldv ), rv = v1 + FPMUL( diff, rdv );
	int lz = z1, rz = z1;
	int iy1 = y1 >> FPP, iy2 = y2 >> FPP, iy3 = y3 >> FPP;
	// Calculate span delta's
	int slx = lx + ldx * (iy2 - iy1);
	int slu = lu + ldu * (iy2 - iy1);
	int slv = lv + ldv * (iy2 - iy1);
	int slz = lz + ldz * (iy2 - iy1);
	int srx = rx + rdx * (iy2 - iy1);
	int sru = ru + rdu * (iy2 - iy1);
	int srv = rv + rdv * (iy2 - iy1);
	int srz = rz + rdz * (iy2 - iy1);
	int len = srx - slx;
	if (len < 1) return;
	int dz = FPDIV( srz - slz, len );
	int du = FPDIV( sru - slu, len );
	int dv = FPDIV( srv - slv, len );
	// End of span delta calculations
	PIXEL* dst = m_Dest + iy1 * 320;
	PIXEL* zbf = m_ZBuffer + iy1 * 320;
	PIXEL* src = a_Poly->gettexture()->getbitmap();
	for ( int i = iy1; i < iy2; i++ )
	{
		int len = rx - lx;
		if (len > 0)
		{
			int diff = (FP_ONE - 1) - (lx & (FP_ONE - 1));
			int cz = lz;
			int cu = lu + FPMUL( diff, du );
			int cv = lv + FPMUL( diff, dv ); 
			int x1 = lx >> FPP;
			int x2 = rx >> FPP;
			for ( int j = x1; j < x2; j++ ) 
			{
				if (*(zbf + j) < (cz >> 13))
				{
					*(dst + j) = *(src + ((cu >> FPP) & 255) + (((cv >> FPP) & 255) << 8));
					*(zbf + j) = cz >> 13;
				}
				cu += du;
				cv += dv;
				cz += dz;
			}
		}
		lx += ldx; lu += ldu; lv += ldv; lz += ldz;
		rx += rdx; ru += rdu; rv += rdv; rz += rdz;
		dst += 320;
		zbf += 320;
	}
	diff = (FP_ONE - 1) - (y2 & (FP_ONE - 1));
	if (lleft)
	{
		rlen = y3 - y2;
		if (rlen == 0) return; else 
		{
			rdx = FPDIV( x3 - x2, rlen ); 
			rdu = FPDIV( u3 - u2, rlen ); 
			rdv = FPDIV( v3 - v2, rlen ); 
			rdz = FPDIV( z3 - z2, rlen ); 
		}
		rx = x2 + FPMUL( diff, rdx );
		ru = u2 + FPMUL( diff, rdu );
		rv = v2 + FPMUL( diff, rdv );
		rz = z2 + FPMUL( diff, rdz );
	}
	else
	{
		llen = y3 - y2;
		if (llen == 0) return; else 
		{
			ldx = FPDIV( x3 - x2, llen );
			ldu = FPDIV( u3 - u2, llen );
			ldv = FPDIV( v3 - v2, llen );
			ldz = FPDIV( z3 - z2, llen );
		}
		lx = x2 + FPMUL( diff, ldx );
		lu = u2 + FPMUL( diff, ldu );
		lv = v2 + FPMUL( diff, ldv );
		lz = z2 + FPMUL( diff, ldz );
	}
	for ( i = iy2; i < iy3; i++ )
	{
		int len = rx - lx;
		if (len > 0)
		{
			int diff = (FP_ONE - 1) - (lx & (FP_ONE - 1));
			int cz = lz;
			int cu = lu + FPMUL( diff, du );
			int cv = lv + FPMUL( diff, dv ); 
			int x1 = lx >> FPP;
			int x2 = rx >> FPP;
			for ( int j = x1; j < x2; j++ ) 
			{
				if (*(zbf + j) < (cz >> 13))
				{
					*(dst + j) = *(src + ((cu >> FPP) & 255) + (((cv >> FPP) & 255) << 8));
					*(zbf + j) = cz >> 13;
				}
				cu += du;
				cv += dv;
				cz += dz;
			}
		}
		lx += ldx; lu += ldu; lv += ldv; lz += ldz;
		rx += rdx; ru += rdu; rv += rdv; rz += rdz;
		dst += 320;
		zbf += 320;
	}
}

// ------------------------------------------------
// EOF